@@ -5,11 +5,22 @@ import os
 import random
 import re
 import socket
-try:
-    import cStringIO as StringIO
-except ImportError:
-    import io as StringIO
+import subprocess
 import sys
+import warnings
+import hpilo_fw
+
+PY3 = sys.version_info[0] >= 3
+if PY3:
+    import io as StringIO
+    b = lambda x: bytes(x, 'ascii')
+    class Bogus(Exception): pass
+    socket.sslerror = Bogus
+    basestring = str
+else:
+    import cStringIO as StringIO
+    b = lambda x: x
+
 try:
     import ssl
 except ImportError:
@@ -19,7 +30,6 @@ except ImportError:
         @staticmethod
         def wrap_socket(sock, *args, **kwargs):
             return socket.ssl(sock)
-import warnings
 try:
     import xml.etree.cElementTree as etree
 except ImportError:
@@ -28,6 +38,7 @@ except ImportError:
 # Which protocol to use
 ILO_RAW  = 1
 ILO_HTTP = 2
+ILO_LOCAL = 3
 
 _untested = []
 
@@ -53,34 +64,41 @@ class IloWarning(Warning):
 
 class Ilo(object):
     """Represents an iLO/iLO2/iLO3/RILOE II management interface on a
-        specific host. A new connection using the specified login, password
-        and timeout will be made for each API call."""
+        specific host. A new connection using the specified login, password and
+        timeout will be made for each API call. The library will detect which
+        protocol to use, but you can override this by setting protocol to
+        ILO_RAW or ILO_HTTP. Use ILO_LOCAL to avoid using a network connection
+        and use hponcfg instead. Username and password are ignored for ILO_LOCAL
+        connections."""
 
-    XML_HEADER = '<?xml version="1.0"?>\r\n'
+    XML_HEADER = b('<?xml version="1.0"?>\r\n')
     HTTP_HEADER = "POST /ribcl HTTP/1.1\r\nHost: localhost\r\nContent-Length: %d\r\nConnection: Close%s\r\n\r\n"
     HTTP_UPLOAD_HEADER = "POST /cgi-bin/uploadRibclFiles HTTP/1.1\r\nHost: localhost\r\nConnection: Close\r\nContent-Length: %d\r\nContent-Type: multipart/form-data; boundary=%s\r\n\r\n"
     BLOCK_SIZE = 4096
+    hponcfg = "/sbin/hponcfg"
 
-    def __init__(self, hostname, login, password, timeout=60, port=443):
+    def __init__(self, hostname, login=None, password=None, timeout=60, port=443, protocol=None):
         self.hostname = hostname
-        self.login = login
-        self.password = password
+        self.login = login or 'Administrator'
+        self.password = password or 'Password'
         self.timeout  = timeout
         self.debug    = 0
-        self.protocol = None
         self.port     = port
+        self.protocol = protocol
         self.cookie   = None
 
     def __str__(self):
         return "iLO interface of %s" % self.hostname
 
     def _debug(self, level, message):
+        if message.__class__.__name__ == 'bytes':
+            message = message.decode('latin-1')
         if self.debug >= level:
+            sys.stderr.write(re.sub(r'PASSWORD=".*"', 'PASSWORD="********"', message))
             if message.startswith('\r'):
-                sys.stderr.write(re.sub(r'PASSWORD=".*"','PASSWORD="********"', message))
                 sys.stderr.flush()
             else:
-                print >>sys.stderr, re.sub(r'PASSWORD=".*"','PASSWORD="********"', message)
+                sys.stderr.write('\n')
 
     def _request(self, xml, progress=None):
         """Given an ElementTree.Element, serialize it and do the request.
@@ -91,7 +109,7 @@ class Ilo(object):
 
         # Serialize the XML
         if hasattr(etree, 'tostringlist'):
-            xml = "\r\n".join(etree.tostringlist(xml)) + '\r\n'
+            xml = b("\r\n").join(etree.tostringlist(xml)) + b('\r\n')
         else:
             xml = etree.tostring(xml)
 
@@ -120,10 +138,14 @@ class Ilo(object):
             return header, messages
 
     def _detect_protocol(self):
+        # Use hponcfg when 'connecting' to localhost
+        if self.hostname == 'localhost':
+            self.protocol = ILO_LOCAL
+            return
         # Do a bogus request, using the HTTP protocol. If there is no
         # header (see special case in communicate(), we should be using the
         # raw protocol
-        header, data = self._communicate('<RIBCL VERSION="2.0"></RIBCL>', ILO_HTTP)
+        header, data = self._communicate(b('<RIBCL VERSION="2.0"></RIBCL>'), ILO_HTTP)
         if header:
             self.protocol = ILO_HTTP
         else:
@@ -131,20 +153,20 @@ class Ilo(object):
 
     def _upload_file(self, filename, progress):
         firmware = open(filename, 'rb').read()
-        boundary = '------hpiLO3t' + str(random.randint(100000,1000000)) + 'z'
+        boundary = b('------hpiLO3t' + str(random.randint(100000,1000000)) + 'z')
         while boundary in firmware:
-            boundary = '------hpiLO3t' + str(random.randint(100000,1000000)) + 'z'
+            boundary = b('------hpiLO3t' + str(random.randint(100000,1000000)) + 'z')
         parts = [
-            """--%s\r\nContent-Disposition: form-data; name="fileType"\r\n\r\n""" % boundary,
-            """\r\n--%s\r\nContent-Disposition: form-data; name="fwimgfile"; filename="%s"\r\nContent-Type: application/octet-stream\r\n\r\n""" % (boundary, filename),
+            b("--") + boundary + b("""\r\nContent-Disposition: form-data; name="fileType"\r\n\r\n"""),
+            b("\r\n--") + boundary + b('''\r\nContent-Disposition: form-data; name="fwimgfile"; filename="''') + b(filename) + b('''"\r\nContent-Type: application/octet-stream\r\n\r\n'''),
             firmware,
-            """\r\n--%s--\r\n""" % boundary
+            b("\r\n--") + boundary + b("--\r\n"),
         ]
         total_bytes = sum([len(x) for x in parts])
         sock = self._get_socket()
 
-        self._debug(2, self.HTTP_UPLOAD_HEADER % (total_bytes, boundary))
-        sock.write(self.HTTP_UPLOAD_HEADER % (total_bytes, boundary))
+        self._debug(2, self.HTTP_UPLOAD_HEADER % (total_bytes, boundary.decode('ascii')))
+        sock.write(b(self.HTTP_UPLOAD_HEADER % (total_bytes, boundary.decode('ascii'))))
         for part in parts:
             if len(part) < 2048:
                 self._debug(2, part)
@@ -170,7 +192,7 @@ class Ilo(object):
         try:
             while True:
                 d = sock.read()
-                data += d
+                data += d.decode('latin-1')
                 if not d:
                     break
         except socket.sslerror: # Connection closed
@@ -184,7 +206,18 @@ class Ilo(object):
         self._debug(2, "Cookie: %s" % self.cookie)
 
     def _get_socket(self):
-        """Set up an https connection and do an HTTP/raw socket request"""
+        """Set up a subprocess or an https connection and do an HTTP/raw socket request"""
+        if self.protocol == ILO_LOCAL:
+            self._debug(1, "Launching hponcfg")
+            try:
+                sp = subprocess.Popen([self.hponcfg, '--input', '--xmlverbose'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=None)
+            except OSError:
+                e = sys.exc_info()[1]
+                raise IloError("Cannot run %s: %s" % (self.hponcfg, str(e)))
+            sp.write = sp.stdin.write
+            sp.read = sp.stdout.read
+            return sp
+
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         sock.settimeout(self.timeout)
         self._debug(1, "Connecting to %s:%d" % (self.hostname, self.port))
@@ -214,14 +247,15 @@ class Ilo(object):
 
         if protocol == ILO_HTTP:
             self._debug(2, http_header)
-            sock.write(http_header)
+            sock.write(b(http_header))
 
         self._debug(2, self.XML_HEADER + xml)
 
         # XML header and data need to arrive in 2 distinct packets
-        sock.write(self.XML_HEADER)
-        if '$EMBED' in xml:
-            pre, name, post = re.compile(r'(.*)\$EMBED:(.*)\$(.*)', re.DOTALL).match(xml).groups()
+        if self.protocol != ILO_LOCAL:
+            sock.write(self.XML_HEADER)
+        if b('$EMBED') in xml:
+            pre, name, post = re.compile(b(r'(.*)\$EMBED:(.*)\$(.*)'), re.DOTALL).match(xml).groups()
             sock.write(pre)
             sent = 0
             fwlen = os.path.getsize(name)
@@ -243,10 +277,13 @@ class Ilo(object):
             sock.write(xml)
 
         # And grab the data
+        if self.protocol == ILO_LOCAL:
+            # hponcfg doesn't return data until stdin is closed
+            sock.stdin.close()
         data = ''
         try:
             while True:
-                d = sock.read()
+                d = sock.read().decode('latin-1')
                 data += d
                 if not d:
                     break
@@ -267,6 +304,9 @@ class Ilo(object):
 
         self._debug(1, "Received %d bytes" % len(data))
 
+        # Stript out garbage from hponcfg
+        if self.protocol == ILO_LOCAL:
+            data = data[data.find('<'):data.rfind('>')+1]
         # Do we have HTTP?
         header_ = ''
         if protocol == ILO_HTTP and data.startswith('HTTP/1.1 200'):
@@ -381,7 +421,7 @@ class Ilo(object):
     def _element_to_dict(self, element):
         """Returns a dict with tag attributes as items"""
         retval = {}
-        for key, val in element.attrib.iteritems():
+        for key, val in element.attrib.items():
             retval[key.lower()] = self._coerce(val)
         return retval
 
@@ -475,6 +515,11 @@ class Ilo(object):
         return self._control_tag('USER_INFO', 'ADD_USER', elements=elements,
                 attrib={'USER_LOGIN': user_login, 'USER_NAME': user_name, 'PASSWORD': password})
 
+    @untested
+    def ahs_clear_data(self):
+        """FIXME: I have no relevant hardware. Please report sample output"""
+        return self._raw(('RIB_INFO', {'MODE': 'WRITE'}), ('AHS_CLEAR_DATA', {}))
+
     def cert_fqdn(self, use_fqdn):
         """Configure whether to use the fqdn or the short hostname for certificate requests"""
         use_fqdn = str({True: 'Yes', False: 'No'}.get(use_fqdn, use_fqdn))
@@ -523,6 +568,11 @@ class Ilo(object):
         """Reset the iLO to factory default settings"""
         return self._control_tag('RIB_INFO', 'FACTORY_DEFAULTS')
 
+    @untested
+    def get_ahs_status(self):
+        """FIXME: I have no relevant hardware. Please report sample output"""
+        return self._raw(('RIB_INFO', {'MODE': 'READ'}), ('GET_AHS_STATUS', {}))
+
     def get_all_users(self):
         """Get a list of all loginnames"""
         data = self._info_tag2('USER_INFO', 'GET_ALL_USERS', key='value')
@@ -766,7 +816,7 @@ class Ilo(object):
             rbsu_post_ip=None, remote_console_encryption=None, remote_keyboard_model=None,
             terminal_services_port=None, high_performance_mouse=None,
             shared_console_enable=None, shared_console_port=None,
-            remote_console_acquire=None):
+            remote_console_acquire=None, brownout_recovery=None):
         """Modify iLO global settings, only values that are specified will be changed."""
         vars = dict(locals())
         del vars['self']
@@ -846,6 +896,12 @@ class Ilo(object):
         """Power cycle the server"""
         return self._control_tag('SERVER_INFO', 'RESET_SERVER')
 
+    @untested
+    def set_ahs_status(self, status):
+        """Enable or disable AHS logging"""
+        status = {True: 'Enable', False: 'Disable'}[status]
+        return self._control_tag('RIB_INFO', 'SET_AHS_STATUS', attrib={'VALUE': status})
+
     def set_language(self, lang_id):
         """Set the default language"""
         return self._control_tag('RIB_INFO', 'SET_LANGUAGE', attrib={'LANG_ID': lang_id})
@@ -921,10 +977,25 @@ class Ilo(object):
         return self._control_tag('SERVER_INFO', 'UID_CONTROL', attrib={"UID": uid.title()})
 
     def update_rib_firmware(self, filename, progress=None):
-        """Upload new RIB firmware"""
+        """Upload new RIB firmware, use "latest" as filename to automatically
+        download and use the latest firmware"""
         if not self.protocol:
             self._detect_protocol()
 
+        if self.protocol == 'ILO_LOCAL':
+            raise ValueError("Cannot update iLO firmware using hponcfg")
+
+        if filename == 'latest':
+            config = hpilo_fw.config()
+            current = self.get_fw_version()
+            ilo = current['management_processor'].lower()
+            if ilo not in config:
+                raise IloError("Cannot update %s to the latest version automatically" % ilo)
+            if current['firmware_version'] >= config[ilo]['version']:
+                return "Already up-to-date"
+            hpilo_fw.download(ilo)
+            filename = config[ilo]['file']
+
         if progress:
             def progress_(data):
                 if data is None:
